@@ -4,7 +4,7 @@
 #
 # @author: Eric Lapouyade
 #
-"""This module defines base classes for python-textops"""
+"""This module defines base classes for python-textops3"""
 
 import os
 import sys
@@ -15,6 +15,8 @@ from addicted import NoAttrDict, NoAttr
 import string
 import logging
 import pprint
+import chardet
+from collections import Callable, abc
 pp = pprint.PrettyPrinter(indent=4)
 
 
@@ -23,7 +25,7 @@ logger = textops.logger
 def activate_debug():
     """Activate debug logging on console
 
-    This function is useful when playing with python-textops through a python console.
+    This function is useful when playing with python-textops3 through a python console.
     It is not recommended to use this function in a real application : use standard logging
     functions instead.
     """
@@ -148,8 +150,8 @@ class TextOp(object):
     def __repr__(self):
         rops = []
         for op,args,kwargs in self.ops:
-            opargs = [ '%r' % v if isinstance(v,(basestring,int,float)) else v for v in args ]
-            opargs += [ '%s=%r' % (k,v) for k,v in kwargs.items() ]
+            opargs = [ '%r' % v if isinstance(v,(str,int,float)) else v for v in args ]
+            opargs += [ '%s=%r' % (k,v) for k,v in list(kwargs.items()) ]
             rops.append('%s(%s)' % (op,','.join(map(str,opargs))))
         return '.'.join(rops)
 
@@ -179,7 +181,7 @@ class TextOp(object):
             <generator object extend_type_gen at ...>
             >>> def mygen(): yield None
             >>> type(echo(None).g)                              # doctest: +ELLIPSIS
-            <type 'NoneType'>
+            <class 'NoneType'>
         """
         text = self._process()
         return self.make_gen(text)
@@ -204,7 +206,7 @@ class TextOp(object):
     def make_list(cls, text, return_if_none=None):
         if text in [None,NoAttr]:
             return return_if_none
-        elif isinstance(text, basestring):
+        elif isinstance(text, str):
             return text.splitlines()
         elif isinstance(text, (types.GeneratorType, enumerate)):
             return ListExt(text)
@@ -226,7 +228,7 @@ class TextOp(object):
             >>> echo('hello').l
             ['hello']
             >>> type(echo(None).g)
-            <type 'NoneType'>
+            <class 'NoneType'>
         """
         text = self._process()
         return self.make_list(text)
@@ -250,7 +252,7 @@ class TextOp(object):
     def make_string(cls, text, join_str='\n', return_if_none=None):
         if text in [None,NoAttr]:
             return return_if_none
-        elif isinstance(text, (list,types.GeneratorType)):
+        elif isinstance(text, (list,types.GeneratorType,abc.ItemsView,abc.KeysView,abc.ValuesView)):
             return StrExt(join_str.join([ stru(item) for item in text ]))
         return StrExt(text)
 
@@ -271,7 +273,7 @@ class TextOp(object):
             >>> echo(['hello','world']).s
             'hello\nworld'
             >>> type(echo(None).s)
-            <type 'NoneType'>
+            <class 'NoneType'>
         """
         text = self._process()
         return self.make_string(text)
@@ -302,7 +304,7 @@ class TextOp(object):
             >>> echo(['hello','world']).j
             'helloworld'
             >>> type(echo(None).j)
-            <type 'NoneType'>
+            <class 'NoneType'>
         """
         text = self._process()
         return self.make_string(text,join_str='')
@@ -329,10 +331,10 @@ class TextOp(object):
                 return int(float(text))
             except (ValueError, TypeError):
                 return 0
-        if isinstance(text, basestring):
+        if isinstance(text, str):
             return _to_int(text)
         elif isinstance(text,(list,tuple)):
-            return map(_to_int,text)
+            return list(map(_to_int,text))
         return text
 
     @property
@@ -358,10 +360,10 @@ class TextOp(object):
                 return float(text)
             except (ValueError, TypeError):
                 return 0.0
-        if isinstance(text, basestring):
+        if isinstance(text, str):
             return _to_float(text)
         elif isinstance(text, (list,tuple)):
-            return map(_to_float,text)
+            return list(map(_to_float,text))
         return text
 
     @property
@@ -430,11 +432,11 @@ class TextOp(object):
         ...         f ... :val6
         ...     g:val7
         ... f: val8'''
-        >>> print parse_indented(s).r
+        >>> print(parse_indented(s).r)
         {'a': 'val1', 'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'}, 'f': 'val8'}
-        >>> print parse_indented(s).pp
+        >>> print(parse_indented(s).pp)
         {   'a': 'val1',
-            'b': {   'c': 'val3', 'd': {   'e': 'val5', 'f': 'val6'}, 'g': 'val7'},
+            'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'},
             'f': 'val8'}
         """
         text = self._process()
@@ -449,10 +451,10 @@ class TextOp(object):
     def _tolist(cls,text):
         if isinstance(text, str):
             return str.splitlines(text)
-        elif isinstance(text, unicode):
-            return unicode.splitlines(text)
+        elif isinstance(text, bytes):
+            return bytes.splitlines(text)
         elif isinstance(text, dict):
-            return text.iteritems()
+            return iter(text.items())
         elif isinstance(text, (int,float)):
             return [text]
         return text
@@ -461,8 +463,8 @@ class TextOp(object):
     def _tosublist(cls,text):
         if isinstance(text, str):
             return [ [line] for line in str.splitlines(text) ]
-        elif isinstance(text, unicode):
-            return [ [line] for line in unicode.splitlines(text) ]
+        elif isinstance(text, bytes):
+            return [ [line] for line in bytes.splitlines(text) ]
         elif isinstance(text, (int,float,dict)):
             return [[text]]
         elif isinstance(text, list) and (not text or not isinstance(text[0],list)):
@@ -471,7 +473,7 @@ class TextOp(object):
 
     @classmethod
     def _tostr(cls,text):
-        if not isinstance(text, basestring):
+        if not isinstance(text, str):
             return '\n'.join(text)
         return text
 
@@ -490,13 +492,13 @@ class TextOp(object):
         return lst.__getitem__(item)
 
 def extend_type(obj):
-    if isinstance(obj,unicode):
-        if not isinstance(obj,UnicodeExt):
-            return UnicodeExt(obj)
-    elif isinstance(obj,basestring):
+    if isinstance(obj,bytes):
+        if not isinstance(obj,BytesExt):
+            return BytesExt(obj)
+    elif isinstance(obj,str):
         if not isinstance(obj,StrExt):
             return StrExt(obj)
-    elif isinstance(obj,(list,tuple)):
+    elif isinstance(obj,(list,tuple,abc.ItemsView,abc.KeysView,abc.ValuesView)):
         if not isinstance(obj,ListExt):
             return ListExt(obj)
     elif isinstance(obj,dict):
@@ -511,8 +513,8 @@ def extend_type_gen(obj):
         yield extend_type(i)
 
 def stru(text):
-    if isinstance(text, unicode):
-        text = text.encode('utf-8','replace')
+    if isinstance(text, bytes):
+        text = text.decode('utf-8','replace')
     else:
         text = str(text)
     return text
@@ -539,7 +541,7 @@ class WrapOpStr(TextOp):
     @classmethod
     def op(cls, text, *args,**kwargs):
         args = list(args)
-        if isinstance(text, basestring):
+        if isinstance(text, str):
             args.insert(cls.input_argn,text)
             return cls.fn(*args,**kwargs)
         else:
@@ -635,11 +637,11 @@ class DebugText(object):
         if isinstance(self.text, str):
             lines = str.splitlines(self.text)
             begin,sep,end = '','',''
-        elif isinstance(self.text, unicode):
-            lines = unicode.splitlines(self.text)
+        elif isinstance(self.text, str):
+            lines = str.splitlines(self.text)
             begin,sep,end = '','',''
         elif isinstance(self.text, dict):
-            lines = [ '%s : %s' % item for item in self.text.iteritems() ]
+            lines = [ '%s : %s' % item for item in self.text.items() ]
             begin,sep,end = '{',',','}'
         elif isinstance(self.text,(types.GeneratorType,list)):
             lines = self.text
@@ -660,7 +662,7 @@ class DebugText(object):
         return out
 
 def get_attribute_or_textop(obj,name):
-    if name[:1]=='_':
+    if name.startswith('__'):
         return object.__getattribute__(obj,name)
     op_cls = getattr(textops.ops,name,None)
     if op_cls and isinstance(op_cls,type) and issubclass(op_cls,TextOp):
@@ -682,12 +684,12 @@ def get_attribute_or_textop(obj,name):
         except AttributeError:
             if isinstance(obj,DictExt):
                 return NoAttr
-            elif isinstance(obj,(StrExt,UnicodeExt)):
+            elif isinstance(obj,(StrExt,BytesExt)):
                 raise
             def fn(*args,**kwargs):
-                return map(lambda s:getattr(str if isinstance(s,str) else unicode, name)(s,*args,**kwargs),obj)
+                return [getattr(str if isinstance(s,str) else str, name)(s,*args,**kwargs) for s in obj]
 
-    if not callable(fn):
+    if not isinstance(fn, Callable):
         return fn
 
     def wrapper(*args, **kwargs):
@@ -701,18 +703,18 @@ def get_attribute_or_textop(obj,name):
         return extend_type(result)
     return wrapper
 
-class UnicodeExt(unicode):
-    """Extend Unicode class to gain access to textops as attributes
+class BytesExt(bytes):
+    """Extend bytes class to gain access to textops as attributes
 
     Examples:
 
-        >>> u'normal unicode'.cut()
+        >>> b'normal bytes'.cut()
         Traceback (most recent call last):
             ...
-        AttributeError: 'unicode' object has no attribute 'cut'
+        AttributeError: 'bytes' object has no attribute 'cut'
 
-        >>> UnicodeExt('extended unicode').cut()
-        [u'extended', u'unicode']
+        >>> BytesExt(b'extended bytes').cut()
+        [b'extended', b'bytes']
     """
     def __getattribute__(self, name):
         return get_attribute_or_textop(self,name)
@@ -721,25 +723,23 @@ class UnicodeExt(unicode):
         """ Convert to ListExt object """
         return ListExt([self])
     def __getslice__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__getslice__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__getslice__(*args, **kwargs))
     def __getitem__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__getitem__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__getitem__(*args, **kwargs))
     def __add__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__add__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__add__(*args, **kwargs))
     def __mul__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__mul__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__mul__(*args, **kwargs))
     def __rmul__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__rmul__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__rmul__(*args, **kwargs))
     def __mod__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__mod__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__mod__(*args, **kwargs))
     def __rmod__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__rmod__(*args, **kwargs))
+        return extend_type(super(BytesExt, self).__rmod__(*args, **kwargs))
     def __format__(self,*args, **kwargs):
-        return extend_type(super(UnicodeExt, self).__format__(*args, **kwargs))
-    def __str__(self):
-        return self.encode('utf-8')    
+        return extend_type(super(BytesExt, self).__format__(*args, **kwargs))
     def _ipython_display_(self):
-        print self.__repr__()
+        print(self.__repr__())
 
 class StrExt(str):
     """Extend str class to gain access to textops as attributes
@@ -777,12 +777,12 @@ class StrExt(str):
     def __format__(self,*args, **kwargs):
         return extend_type(super(StrExt, self).__format__(*args, **kwargs))
     def _ipython_display_(self):
-        print self.__repr__()
+        print(self.__repr__())
 
 class ListExt(list):
     """Extend list class to gain access to textops as attributes
 
-    In addition, all list items (dict, list, str, unicode) are extended on-the-fly when accessed
+    In addition, all list items (dict, list, str, bytes) are extended on-the-fly when accessed
 
     Examples:
 
@@ -822,7 +822,7 @@ class ListExt(list):
     def __iter__(self):
         return ListExtIterator(self)
     def _ipython_display_(self):
-        print self.__repr__()
+        print(self.__repr__())
 
 class ListExtIterator(object):
     def __init__(self, obj):
@@ -831,7 +831,7 @@ class ListExtIterator(object):
         self.len = len(obj)
     def __iter__(self):
         return self
-    def next(self):
+    def __next__(self):
         self.index += 1
         if self.index < self.len:
             # Will call extend_type() because __getitem__ do it :
@@ -845,7 +845,7 @@ class DictExt(NoAttrDict):
     New features are :
 
         * Access to textops operations with attribute notation
-        * All dict values (dict, list, str, unicode) are extended on-the-fly when accessed
+        * All dict values (dict, list, str, bytes) are extended on-the-fly when accessed
         * Access to dict values with attribute notation
         * Add a key:value in the dict with attribute notation (one level at a time)
         * Returns NoAttr object when a key is not in the Dict
@@ -862,13 +862,13 @@ class DictExt(NoAttrDict):
         >>> {'a':1,'b':2}.items().grep('a')
         Traceback (most recent call last):
             ...
-        AttributeError: 'list' object has no attribute 'grep'
+        AttributeError: 'dict_items' object has no attribute 'grep'
 
         >>> DictExt({'a':1,'b':2}).items().grep('a')
         [['a', 1]]
 
         >>> d = DictExt({ 'this' : { 'is' : { 'a' : {'very deep' : { 'dict' : 'yes it is'}}}}})
-        >>> print d.this['is'].a['very deep'].dict
+        >>> print(d.this['is'].a['very deep'].dict)
         yes it is
         >>> d.not_a_valid_key
         NoAttr
@@ -879,29 +879,40 @@ class DictExt(NoAttrDict):
         >>> for obj in d.not_a_valid_key.objects:
         ...     do_things(obj)
         ... else:
-        ...     print 'no object'
+        ...     print('no object')
         no object
 
         >>> d = DictExt()
         >>> d.a = DictExt()
         >>> d.a.b = 'this is my logging data'
-        >>> print d
+        >>> print(d)
         {'a': {'b': 'this is my logging data'}}
 
         >>> d = { 'mykey' : 'myval' }
         >>> d['mykey']
         'myval'
         >>> type(d['mykey'])
-        <type 'str'>
+        <class 'str'>
         >>> d = DictExt(d)
         >>> d['mykey']
         'myval'
         >>> type(d['mykey'])
         <class 'textops.base.StrExt'>
+
+        >>> d=DictExt()
+        >>> d[0]=[]
+        >>> d
+        {0: []}
+        >>> d[0].append(3)
+        >>> d
+        {0: [3]}
+        >>> type(d[0])
+        <class 'textops.base.ListExt'>
+
     """
     def __getattribute__(self, name):
-        if dict.has_key(self,name):
-            return self[name]
+        if dict.__contains__(self,name):
+            return dict.__getitem__(self,name)
         return get_attribute_or_textop(self,name)
 
     @property
@@ -919,9 +930,9 @@ class DictExt(NoAttrDict):
             >>> s = '''soft:textops
             ... count:32591'''
             >>> s | parse_indented()
-            {'count': '32591', 'soft': 'textops'}
+            {'soft': 'textops', 'count': '32591'}
             >>> s | parse_indented().amend(date='2015-11-19')
-            {'count': '32591', 'date': '2015-11-19', 'soft': 'textops'}
+            {'soft': 'textops', 'count': '32591', 'date': '2015-11-19'}
         """
         return DictExt(self,*args, **kwargs)
     def render(self,format_string,defvalue='-'):
@@ -955,7 +966,7 @@ class DefaultDict(dict):
         try:
             return super(DefaultDict,self).__getitem__(key)
         except KeyError:
-            if callable(self.defvalue):
+            if isinstance(self.defvalue, Callable):
                 return self.defvalue(key)
             return self.defvalue
 
@@ -967,7 +978,7 @@ class DefaultList(list):
         try:
             return super(DefaultList,self).__getitem__(key)
         except IndexError:
-            if callable(self.defvalue):
+            if isinstance(self.defvalue, Callable):
                 return self.defvalue(key)
             return self.defvalue
 
@@ -986,7 +997,7 @@ def dictmerge(*dict_args):
     Examples:
 
         >>> dictmerge({'a':1,'b':2},{'b':3,'c':4})
-        {'a': 1, 'c': 4, 'b': 3}
+        {'a': 1, 'b': 3, 'c': 4}
 
     """
     result = {}
@@ -1046,6 +1057,61 @@ def eformat(format_str,lst,dct,defvalue='-'):
     """
     return vformat(format_str,DefaultList(defvalue,lst),DefaultDict(defvalue,dct))
 
+
+def decode_bytes(byte_string, encoding=None):
+    """Returns a decoded string from a byte string.
+    It can test several encodings, will try to detect encoding if it does not
+    work and have a ascii/'replace' fallback.
+    One can give standard python3 string, in which case, no action is done
+
+    Args:
+
+        byte_string (bytes): string to decode to python3 standard string
+        encoding (str, list, tuple): ending(s) to use. if a list/tuple is given,
+            will use the first that does not raise exception
+            default : 'utf-8'
+
+    Examples:
+
+        >>> b = 'éric'.encode('utf-16')
+        >>> print(b.hex())
+        fffee900720069006300
+        >>> decode_bytes(b)
+        'éric'
+        >>> decode_bytes(b,'utf-16')
+        'éric'
+        >>> decode_bytes(b,('utf-16','utf-32'))
+        'éric'
+
+    Returns:
+
+        a python3 standard string
+    """
+    if not isinstance(byte_string, bytes):
+        return byte_string
+    if not encoding:
+        encoding = ('utf-8',)
+    if isinstance(encoding, str):
+        encoding = (encoding,)
+    for enc in encoding:
+        try:
+            s = byte_string.decode(enc)
+            return s
+        except UnicodeDecodeError:
+            pass
+    # If given encoding does not work try to detect it :
+    detected_encoding = chardet.detect(byte_string)
+    if detected_encoding['confidence'] >= 0.5:
+        try:
+            s = byte_string.decode(detected_encoding['encoding'])
+            return s
+        except UnicodeDecodeError:
+            pass
+    # Fallback : ascii + replace unkown chars
+    s = byte_string.decode('ascii','replace')
+    return s
+
+
 # NoAttr.as_list must return an empty ListExt() not a simple list
 # This customize noattr module for textops.
 from noattr import NoAttrType
