@@ -8,10 +8,12 @@
 
 from textops import TextOp, NoAttr, dformat, pp, stru
 import textops
+import types
 import string
 import re
 import copy
 from datetime import datetime
+import collections
 
 logger = textops.logger
 
@@ -76,10 +78,10 @@ class mgrep(TextOp):
         ... 'warnings' : r'^warn',
         ... 'infos' : r'^info',
         ... })
-        >>> print t                                         #doctest: +NORMALIZE_WHITESPACE
-        {'infos': ['info 1', 'info 2'],
-        'errors': ['error 1', 'error 2'],
-        'warnings': ['warning 1', 'warning 2']}
+        >>> print(t)  #doctest: +NORMALIZE_WHITESPACE
+        {'errors': ['error 1', 'error 2'],
+        'warnings': ['warning 1', 'warning 2'],
+        'infos': ['info 1', 'info 2']}
 
         >>> s = '''
         ... Disk states
@@ -97,9 +99,9 @@ class mgrep(TextOp):
         ... 'states' : r'^state:',
         ... 'fss' : r'^fs:',
         ... })
-        >>> print t                                         #doctest: +NORMALIZE_WHITESPACE
-        {'states': ['state: good', 'state: failed'],
-        'disks': ['name: c1t0d0s0', 'name: c1t0d0s4'],
+        >>> print(t)  #doctest: +NORMALIZE_WHITESPACE
+        {'disks': ['name: c1t0d0s0', 'name: c1t0d0s4'],
+        'states': ['state: good', 'state: failed'],
         'fss': ['fs: /', 'fs: /home']}
         >>> dict(zip(t.disks.cutre(': *',1),zip(t.states.cutre(': *',1),t.fss.cutre(': *',1))))
         {'c1t0d0s0': ('good', '/'), 'c1t0d0s4': ('failed', '/home')}
@@ -108,14 +110,14 @@ class mgrep(TextOp):
     reverse = False
     @classmethod
     def op(cls,text,patterns_dict,key = None, *args,**kwargs):
-        for k,pattern in patterns_dict.items():
-            if isinstance(pattern,basestring):
+        for k,pattern in list(patterns_dict.items()):
+            if isinstance(pattern,str):
                 patterns_dict[k] = re.compile(pattern,cls.flags)
         dct = {}
         for line in cls._tolist(text):
-            for k,regex in patterns_dict.items():
+            for k,regex in list(patterns_dict.items()):
                 try:
-                    if isinstance(line,basestring):
+                    if isinstance(line,str):
                         if bool(regex.search(stru(line))) != cls.reverse:  # kind of XOR with cls.reverse
                             dct.setdefault(k,[]).append(line)
                     elif key is None:
@@ -172,7 +174,7 @@ class mgrepv(mgrep):
         ... 'not_errors' : r'^err',
         ... 'not_warnings' : r'^warn',
         ... })
-        >>> print t                                         #doctest: +NORMALIZE_WHITESPACE
+        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
         {'not_warnings': ['error 1', 'error 2'], 'not_errors': ['warning 1', 'warning 2']}
     """
     reverse = True
@@ -200,14 +202,14 @@ class mgrepvi(mgrepv):
         ... 'not_errors' : r'^err',
         ... 'not_warnings' : r'^warn',
         ... })
-        >>> print t                                         #doctest: +NORMALIZE_WHITESPACE
+        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
         {'not_warnings': ['error 1', 'WARNING 1', 'ERROR 2'],
         'not_errors': ['WARNING 1', 'warning 2', 'ERROR 2']}
         >>> t = logs | mgrepvi({
         ... 'not_errors' : r'^err',
         ... 'not_warnings' : r'^warn',
         ... })
-        >>> print t                                         #doctest: +NORMALIZE_WHITESPACE
+        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
         {'not_warnings': ['error 1', 'ERROR 2'], 'not_errors': ['WARNING 1', 'warning 2']}
     """
     flags = re.IGNORECASE
@@ -251,19 +253,19 @@ class sgrep(TextOp):
         ... info 2
         ... '''
         >>> t = logs | sgrep(('^err','^warn'))
-        >>> print t                                         #doctest: +NORMALIZE_WHITESPACE
+        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
         [['error 1', 'error 2'], ['warning 1', 'warning 2'], ['', 'info 1', 'info 2']]
     """
     flags = 0
     reverse = False
     @classmethod
     def op(cls,text,patterns,key = None, *args,**kwargs):
-        patterns = [ re.compile(pattern,cls.flags) if isinstance(pattern,basestring) else pattern for pattern in patterns ]
-        lst = [ [] for i in xrange(len(patterns)+1) ] # surtout pas faire [] * (len(patterns)+1)
+        patterns = [ re.compile(pattern,cls.flags) if isinstance(pattern,str) else pattern for pattern in patterns ]
+        lst = [ [] for i in range(len(patterns)+1) ] # surtout pas faire [] * (len(patterns)+1)
         for line in cls._tolist(text):
             for i,regex in enumerate(patterns):
                 try:
-                    if isinstance(line,basestring):
+                    if isinstance(line,str):
                         if bool(regex.search(stru(line))) != cls.reverse:  # kind of XOR with cls.reverse
                             lst[i].append(line)
                             break
@@ -323,7 +325,7 @@ class parseg(TextOp):
     ignore_case = False
     @classmethod
     def op(cls,text, pattern, *args,**kwargs):
-        if isinstance(pattern,basestring):
+        if isinstance(pattern,str):
             pattern = re.compile(pattern, re.I if cls.ignore_case else 0)
         out = []
         for line in cls._tolist(text):
@@ -346,8 +348,8 @@ class parsegi(parseg):
         ... Notice: textops rocks
         ... Warning: Python must be used without moderation'''
         >>> s | parsegi(r'(?P<level>error|warning):\s*(?P<msg>.*)')         #doctest: +NORMALIZE_WHITESPACE
-        [{'msg': 'System will reboot', 'level': 'Error'},
-        {'msg': 'Python must be used without moderation', 'level': 'Warning'}]
+        [{'level': 'Error', 'msg': 'System will reboot'},
+        {'level': 'Warning', 'msg': 'Python must be used without moderation'}]
     """
     ignore_case = True
 
@@ -376,7 +378,7 @@ class parsek(TextOp):
     ignore_case = False
     @classmethod
     def op(cls,text, pattern, key_name = 'key', key_update = None, *args,**kwargs):
-        if isinstance(pattern,basestring):
+        if isinstance(pattern,str):
             pattern = re.compile(pattern, re.I if cls.ignore_case else 0)
         out = []
         for line in cls._tolist(text):
@@ -440,19 +442,19 @@ class parsekv(TextOp):
         ... first name: Eric
         ... country: France'''
         >>> s | parsekv(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
-        {'country': {'val': 'France', 'key': 'country'},
-        'first_name': {'val': 'Eric', 'key': 'first name'},
-        'name': {'val': 'Lapouyade', 'key': 'name'}}
+        {'name': {'key': 'name', 'val': 'Lapouyade'},
+        'first_name': {'key': 'first name', 'val': 'Eric'},
+        'country': {'key': 'country', 'val': 'France'}}
         >>> s | parsekv(r'(?P<item>.*):\s*(?P<val>.*)','item',str.upper)         #doctest: +NORMALIZE_WHITESPACE
-        {'FIRST NAME': {'item': 'first name', 'val': 'Eric'},
-        'NAME': {'item': 'name', 'val': 'Lapouyade'},
+        {'NAME': {'item': 'name', 'val': 'Lapouyade'},
+        'FIRST NAME': {'item': 'first name', 'val': 'Eric'},
         'COUNTRY': {'item': 'country', 'val': 'France'}}
         >>> s | parsekv(r'(?P<key>.*):\s*(?P<val>.*)',key_update=0)         #doctest: +NORMALIZE_WHITESPACE
-        {'first name': {'val': 'Eric', 'key': 'first name'},
-        'name': {'val': 'Lapouyade', 'key': 'name'},
-        'country': {'val': 'France', 'key': 'country'}}
+        {'name': {'key': 'name', 'val': 'Lapouyade'},
+        'first name': {'key': 'first name', 'val': 'Eric'},
+        'country': {'key': 'country', 'val': 'France'}}
         >>> s | parsekv(r'(?P<key>.*):\s*(?P<val>.*)',val_name='val')         #doctest: +NORMALIZE_WHITESPACE
-        {'country': 'France', 'first_name': 'Eric', 'name': 'Lapouyade'}
+        {'name': 'Lapouyade', 'first_name': 'Eric', 'country': 'France'}
     """
     ignore_case = False
     val_name = None
@@ -460,23 +462,29 @@ class parsekv(TextOp):
     def op(cls,text, pattern, key_name = 'key', key_update = None, val_name = None, *args,**kwargs):
         if val_name is None:
             val_name = cls.val_name
-        if isinstance(pattern,basestring):
+        if isinstance(pattern,str):
             pattern = re.compile(pattern, re.I if cls.ignore_case else 0)
-        out = {}
-        for line in cls._tolist(text):
-            m = pattern.match(line)
-            if m:
-                dct = m.groupdict()
-                key = dct.get(key_name)
-                if key:
-                    if key_update is None:
-                        key_norm = index_normalize(key)
-                    elif callable(key_update):
-                        key_norm = key_update(key)
-                    else:
-                        key_norm = key
-                    out.update({ key_norm : dct if val_name is None else dct[val_name]})
-        return out
+
+        def _op(text):
+            out = {}
+            for line in cls._tolist(text):
+                m = pattern.match(line)
+                if m:
+                    dct = m.groupdict()
+                    key = dct.get(key_name)
+                    if key:
+                        if key_update is None:
+                            key_norm = index_normalize(key)
+                        elif isinstance(key_update, collections.Callable):
+                            key_norm = key_update(key)
+                        else:
+                            key_norm = key
+                        out.update({ key_norm : dct if val_name is None else dct[val_name]})
+            return out
+
+        if isinstance(text,(list,types.GeneratorType)):
+            return [ _op(item) for item in text ]
+        return _op(text)
 
 class parsekvi(parsekv):
     r"""Find all occurrences of one pattern (case insensitive), returns a dict of groupdicts
@@ -501,7 +509,7 @@ class parsekvi(parsekv):
         ... first name: Eric
         ... country: France'''
         >>> s | parsekvi(r'(?P<key>NAME):\s*(?P<val>.*)')
-        {'name': {'val': 'Lapouyade', 'key': 'name'}}
+        {'name': {'key': 'name', 'val': 'Lapouyade'}}
     """
     ignore_case = True
 
@@ -509,6 +517,7 @@ class keyval(parsekv):
     r"""Return a dictionnay where keys and values are taken from the pattern specify
 
     It is a shortcut for :class:`textops.parsekv` with val_name='val'
+    The input can be a string or a list of strings.
 
     Args:
         pattern (str): a regular expression string.
@@ -521,14 +530,22 @@ class keyval(parsekv):
             the value at the key ``val_name`. (by default, None means 'val')
 
     Returns:
-        dict: A dict of key:val from the matched pattern groupdict
+        dict: A dict of key:val from the matched pattern groupdict or a list of dicts if the input is a list of strings
 
     Examples:
         >>> s = '''name: Lapouyade
         ... first name: Eric
         ... country: France'''
         >>> s | keyval(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
-        {'country': 'France', 'first_name': 'Eric', 'name': 'Lapouyade'}
+        {'name': 'Lapouyade', 'first_name': 'Eric', 'country': 'France'}
+
+        >>> s = [ '''name: Lapouyade
+        ... first name: Eric ''',
+        ... '''name: Python
+        ... first name: Guido''' ]
+        >>> s | keyval(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
+        [{'name': 'Lapouyade', 'first_name': 'Eric '}, {'name': 'Python', 'first_name': 'Guido'}]
+
     """
     val_name = 'val'
 
@@ -555,7 +572,7 @@ class keyvali(keyval):
         ... first name IS Eric
         ... country IS France'''
         >>> s | keyvali(r'(?P<key>.*) is (?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
-        {'country': 'France', 'first_name': 'Eric', 'name': 'Lapouyade'}
+        {'name': 'Lapouyade', 'first_name': 'Eric', 'country': 'France'}
     """
     ignore_case = True
 
@@ -587,7 +604,7 @@ class find_pattern(TextOp):
 
     @classmethod
     def op(cls,text, pattern, *args,**kwargs):
-        if isinstance(pattern,basestring):
+        if isinstance(pattern,str):
             pattern = re.compile(pattern, re.M | (re.I if cls.ignore_case else 0))
         text = cls._tostr(text)
         m = pattern.search(text)
@@ -643,14 +660,14 @@ class find_patterns(TextOp):
         ... 'format':r'^Format:\s*(?P<format>.*)',
         ... })
         >>> r
-        {'version': {'major': '1', 'build': '3', 'minor': '2'}, 'format': 'json'}
+        {'version': {'major': '1', 'minor': '2', 'build': '3'}, 'format': 'json'}
         >>> r.version.major
         '1'
         >>> s | find_patterns({
         ... 'version':r'^Version:\s*(\d+)\.(\d+)\.(\d+)',
         ... 'format':r'^Format:\s*(.*)',
         ... })
-        {'version': {'group1': '2', 'group0': '1', 'group2': '3'}, 'format': 'json'}
+        {'version': {'group0': '1', 'group1': '2', 'group2': '3'}, 'format': 'json'}
         >>> s | find_patterns({'version':r'^version:\s*(.*)'}) # lowercase 'version' : no match
         {}
         >>> s = '''creation: 2015-10-14
@@ -660,7 +677,7 @@ class find_patterns(TextOp):
         ['2015-11-16', '2015-11-17', '2015-10-14']
         >>> s | find_patterns([r'^update:\s*(?P<year>.*)-(?P<month>.*)-(?P<day>.*)',
         ... r'^access:\s*(.*)', r'^creation:\s*(.*)'])
-        [{'month': '11', 'day': '16', 'year': '2015'}, '2015-11-17', '2015-10-14']
+        [{'year': '2015', 'month': '11', 'day': '16'}, '2015-11-17', '2015-10-14']
     """
     stop_when_found = False
     ignore_case = False
@@ -670,18 +687,18 @@ class find_patterns(TextOp):
         out = []
         text = cls._tostr(text)
         if isinstance(patterns, dict):
-            patterns_list = patterns.items()
+            patterns_list = list(patterns.items())
         else:
             patterns_list = enumerate(patterns)
         for attr,pattern in patterns_list:
-            if isinstance(pattern,basestring):
+            if isinstance(pattern,str):
                 pattern = re.compile(pattern, re.M | (re.I if cls.ignore_case else 0))
             if pattern:
                 m = pattern.search(text)
                 if m :
                     tmp_groupdict = m.groupdict() or dict([('group%s' % k,v) for k,v in enumerate(m.groups())])
                     groupdict = {}
-                    for grp, val in tmp_groupdict.items():
+                    for grp, val in list(tmp_groupdict.items()):
                         if grp[:3] == 'INT':
                             try:
                                 groupdict[grp[3:]] = int(val)
@@ -747,7 +764,7 @@ class find_first_pattern(find_patterns):
         >>> s | find_first_pattern([r'^UPDATE:\s*(.*)'])
         NoAttr
         >>> s | find_first_pattern([r'^update:\s*(?P<year>.*)-(?P<month>.*)-(?P<day>.*)'])
-        {'year': '2015', 'day': '16', 'month': '11'}
+        {'year': '2015', 'month': '11', 'day': '16'}
     """
     stop_when_found = True
 
@@ -828,7 +845,7 @@ class parse_indented(TextOp):
                         indent -= 1
                         dct = indent_node.get(indent)
                     indent_level = indent
-                    for ik in indent_node.keys():
+                    for ik in list(indent_node.keys()):
                         if ik > indent:
                             del indent_node[ik]
                 elif indent > indent_level:
@@ -844,10 +861,10 @@ class parse_indented(TextOp):
                     if isinstance(prev_v,dict):
                         dct[k]=[prev_v,{}]
                         dct = dct[k][-1]
-                    elif isinstance(prev_v,basestring):
+                    elif isinstance(prev_v,str):
                         dct[k]=[prev_v,v]
                     else:
-                        if isinstance(prev_v[0],basestring):
+                        if isinstance(prev_v[0],str):
                             dct[k].append(v)
                         else:
                             dct[k].append({})
@@ -863,7 +880,14 @@ class parse_smart(TextOp):
 
     It looks for key/value patterns, store found values in a dictionary.
     It tries to respect indents by creating sub-dictionaries.
-    The keys are normalized (only keep ``A-Za-z0-9_``), the values are stripped.
+    The keys are normalized (only keep ``A-Za-z0-9_``, the original key value
+    is stored into the inner dict under the '_original_key' key),
+    the values are stripped.
+
+    Args:
+        key_filter (func): a function that will receive a key before
+        normalization and will return a new key string. The could be useful
+        when a chapter title is too long. (Defaut : no filtering)
 
     Returns:
         dict: structured keys:values
@@ -897,37 +921,45 @@ class parse_smart(TextOp):
         ... DASD DEVICE
         ... '''
         >>> parsed = s >> parse_smart()
-        >>> print parsed.pretty()
+        >>> print(parsed.pretty())
         {   'class': 'H',
             'date_time': 'Wed Dec  2 09:51:17 NFT 2015',
             'description': ['DISK OPERATION ERROR'],
             'location': 'U78AA.001.WZSHM0M-P1-C6-T1-W201400A0B8292A18-L13000000000000',
-            'machine_id': {   'machine_id': '00F7B0114C00', 'node_id': 'xvio6'},
+            'machine_id': {   '_original_key': 'Machine Id',
+                              'machine_id': '00F7B0114C00',
+                              'node_id': 'xvio6'},
             'probable_causes': ['DASD DEVICE'],
             'resource_type': 'mpioapdisk',
             'sequence_number': '156637',
-            'type': {   'resource_name': {   'resource_class': 'disk',
+            'type': {   '_original_key': 'Type',
+                        'resource_name': {   '_original_key': 'Resource Name',
+                                             'resource_class': 'disk',
                                              'resource_name': 'hdisk21'},
                         'type': 'PERM',
                         'wpar': 'Global'},
-            'vpd': {   'device_specific_z0': '0000053245004032',
+            'vpd': {   '_original_key': 'VPD',
+                       'device_specific_z0': '0000053245004032',
                        'device_specific_z1': '',
                        'machine_type_and_model': '1815      FAStT',
                        'manufacturer': 'IBM',
                        'ros_level_and_id': '30393134',
                        'serial_number': ''}}
-        >>> print parsed.vpd.device_specific_z0
+        >>> print(parsed.vpd.device_specific_z0)
         0000053245004032
     """
     @classmethod
-    def op(cls, text, *args,**kwargs):
+    def op(cls, text, key_filter=None, *args,**kwargs):
         sep = r'[^a-zA-Z0-9_()\.-]{2,}|[:=]|\.{2,}|[_-]{3,}'
         indent_level = 0
         out = {}
         indent_node = {indent_level:out}
         dct = out
         prev_k = None
+        prev_original_k = None
         block_step = 1
+        if key_filter is None:
+            key_filter=lambda x:x
         # parse the text
         for line in cls._tolist(text):
             if not line.strip():
@@ -936,6 +968,7 @@ class parse_smart(TextOp):
                 m = re.match(r'^(\s*)(\S.*)', line)
                 if m:
                     k,v = (re.split(sep,m.group(2),1) + [''])[:2]
+                    original_k = k
                     v = v.strip()
                     indent = len(m.group(1))
                     #print indent,indent_level,block_step,line,'****',prev_k,k
@@ -947,7 +980,8 @@ class parse_smart(TextOp):
                     elif block_step == 2:
                         if indent == indent_level:
                             block_k = prev_k
-                            prev_k = index_normalize(k)
+                            prev_k = index_normalize(key_filter(k))
+                            prev_original_k = original_k
                             dct[block_k] = [m.group(2)]
                             block_step = 3
                             continue
@@ -956,7 +990,8 @@ class parse_smart(TextOp):
                     elif block_step == 3:
                         if indent == indent_level:
                             dct[block_k].append(m.group(2))
-                            prev_k = index_normalize(k)
+                            prev_original_k = original_k
+                            prev_k = index_normalize(key_filter(k))
                             continue
                         else:
                             block_step = 0
@@ -969,7 +1004,7 @@ class parse_smart(TextOp):
                             indent -= 1
                             dct = indent_node.get(indent)
                         indent_level = indent
-                        for ik in indent_node.keys():
+                        for ik in list(indent_node.keys()):
                             if ik > indent:
                                 del indent_node[ik]
                         if not v:
@@ -977,14 +1012,15 @@ class parse_smart(TextOp):
                     elif indent > indent_level:
                         if prev_k is not None:
                             if prev_v:
-                                dct[prev_k] = {prev_k:prev_v}
+                                dct[prev_k] = {prev_k:prev_v,
+                                               '_original_key':prev_original_k}
                             else:
-                                dct[prev_k] = {}
+                                dct[prev_k] = {'_original_key':prev_original_k}
                             dct = dct[prev_k]
                         indent_node[indent] = dct
                         indent_level = indent
 
-                    k = index_normalize(k)
+                    k = index_normalize(key_filter(k))
                     if k in dct:
                         prev_v = dct[k]
                         if isinstance(prev_v,dict):
@@ -1002,6 +1038,7 @@ class parse_smart(TextOp):
                     else:
                         dct[k]=v
                         prev_k = k
+                        prev_original_k = original_k
                         prev_v = v
         return out
 
@@ -1043,6 +1080,8 @@ class state_pattern(TextOp):
         use an empty string to not change the current state. One can use any string, usually,
         it corresponds to a specific section name of the document to parse where specific
         rules has to be used.
+        if the pattern matches, no more rules are used for the current line except when you specify
+        ``__continue__`` for the goto state. This is useful when you want to apply several rules on the same line.
 
     ``<pattern>``
         is a string or a re.regex to match a line of text.
@@ -1161,8 +1200,8 @@ class state_pattern(TextOp):
         >>> s | state_pattern( (('',None,'(?P<key>.*):(?P<val>.*)','{key}','{val}'),) )
         {'first_name': 'Eric', 'last_name': 'Lapouyade'}
         >>> s | state_pattern( (('',None,'(?P<key>.*):(?P<val>.*)','{key}',None),) ) #doctest: +NORMALIZE_WHITESPACE
-        {'first_name': {'val': 'Eric', 'key': 'first name'},
-        'last_name': {'val': 'Lapouyade', 'key': 'last name'}}
+        {'first_name': {'key': 'first name', 'val': 'Eric'},
+        'last_name': {'key': 'last name', 'val': 'Lapouyade'}}
         >>> s | state_pattern((('',None,'(?P<key>.*):(?P<val>.*)','my.path.{key}','{val}'),))
         {'my': {'path': {'first_name': 'Eric', 'last_name': 'Lapouyade'}}}
 
@@ -1185,8 +1224,8 @@ class state_pattern(TextOp):
         ... ('','section2','^Section 2',None,None),
         ... ('section1', '', '(?P<key>.*)=(?P<val>.*)', 'section1.{key}', '{val}'),
         ... ('section2', '', '(?P<key>.*):(?P<val>.*)', 'section2.{key}', '{val}')) )
-        {'section2': {'first_name': 'Eric', 'last_name': 'Dupont'},
-        'section1': {'email': 'ericdupo@gmail.com'}}
+        {'section1': {'email': 'ericdupo@gmail.com'},
+        'section2': {'first_name': 'Eric', 'last_name': 'Dupont'}}
 
         >>> s = '''
         ... Disk states
@@ -1223,8 +1262,23 @@ class state_pattern(TextOp):
         ... ('disk', '', r'(?P<key>.*):(?P<val>.*)', '>>disk_info.{key}', '{val}'),
         ... ('disk', 'top', r'}', 'disks.{disk_info[name]}', '<disk_info'),
         ... ) )
-        {'disks': {'c1t0d0s0': {'state': 'good', 'fs': '/', 'name': 'c1t0d0s0'},
+        {'disks': {'c1t0d0s0': {'name': 'c1t0d0s0', 'state': 'good', 'fs': '/'},
         'c1t0d0s4': {'fs': '/home', 'name': 'c1t0d0s4'}}}
+
+        >>> s='firstname:Eric lastname=Lapouyade'
+        >>> s | state_pattern((
+        ... ('top','',r'firstname:(?P<val>\S+)','firstname','{val}'),
+        ... ('top','',r'.*lastname=(?P<val>\S+)','lastname','{val}'),
+        ... ))
+        {'firstname': 'Eric'}
+
+        >>> s='firstname:Eric lastname=Lapouyade'
+        >>> s | state_pattern((
+        ... ('top','__continue__',r'firstname:(?P<val>\S+)','firstname','{val}'),
+        ... ('top','',r'.*lastname=(?P<val>\S+)','lastname','{val}'),
+        ... ))
+        {'firstname': 'Eric', 'lastname': 'Lapouyade'}
+
     """
 
     @classmethod
@@ -1249,11 +1303,11 @@ class state_pattern(TextOp):
             if not ifstate:
                 ifstate = ()
             else:
-                if isinstance(ifstate, basestring):
+                if isinstance(ifstate, str):
                     ifstate = ifstate.split(',')
-            if isinstance(pattern, basestring):
+            if isinstance(pattern, str):
                 pattern = re.compile(pattern,reflags)
-            if isinstance(datapath, basestring):
+            if isinstance(datapath, str):
                 if not datapath:
                     datapath = []
                 else:
@@ -1270,7 +1324,7 @@ class state_pattern(TextOp):
                     if m:
                         g = m.groupdict()
                         if autostrip:
-                            g = dict([ (k,v.strip()) for k,v in m.groupdict().items() ])
+                            g = dict([ (k,v.strip() if isinstance(v,str) else v) for k,v in list(m.groupdict().items()) ])
                         groups_context.update(g,_ifstate=ifstate,_gotostate=gotostate,_state=state)
                         logger.debug('  -> OK')
                         logger.debug('    context = %s',groups_context)
@@ -1302,11 +1356,11 @@ class state_pattern(TextOp):
                                         data[p] = {}
                                     data = data[p]
 
-                            if callable(outfilter):
+                            if isinstance(outfilter, collections.Callable):
                                 g,new_groups_context=outfilter(m,groups_context)
                                 if new_groups_context is not None:
                                     groups_context = new_groups_context
-                            elif isinstance(outfilter, basestring):
+                            elif isinstance(outfilter, str):
                                 if outfilter.startswith('<'):
                                     g = groups_context.get(outfilter[1:].strip(),{})
                                 else:
@@ -1321,6 +1375,8 @@ class state_pattern(TextOp):
                                     data.update(g)
                                 else:
                                     prev_data[p] = g
+                        if gotostate == '__continue__':
+                            continue
                         if gotostate:
                             state = gotostate
                         break
